import calendar
import datetime
import logging
from typing import Optional

import allure
import pytest
from frostfs_testlib import reporter
from frostfs_testlib.resources.error_patterns import OBJECT_NOT_FOUND
from frostfs_testlib.steps.cli.object import get_netmap_netinfo, get_object_from_random_node, head_object
from frostfs_testlib.steps.epoch import get_epoch, wait_for_epochs_align
from frostfs_testlib.steps.http_gate import (
    attr_into_str_header_curl,
    try_to_get_object_and_expect_error,
    upload_via_http_gate_curl,
    verify_object_hash,
)
from frostfs_testlib.storage.dataclasses.object_size import ObjectSize
from frostfs_testlib.storage.dataclasses.wallet import WalletInfo
from frostfs_testlib.testing.cluster_test_base import ClusterTestBase
from frostfs_testlib.utils.file_utils import generate_file

from ....helpers.container_request import REP_2_1_2_PUBLIC, requires_container

logger = logging.getLogger("NeoLogger")
EXPIRATION_TIMESTAMP_HEADER = "__SYSTEM__EXPIRATION_TIMESTAMP"

# TODO: Depreacated. Use EXPIRATION_EPOCH_ATTRIBUTE from testlib
EXPIRATION_EPOCH_HEADER = "__SYSTEM__EXPIRATION_EPOCH"

EXPIRATION_DURATION_HEADER = "__SYSTEM__EXPIRATION_DURATION"
EXPIRATION_EXPIRATION_RFC = "__SYSTEM__EXPIRATION_RFC3339"
SYSTEM_EXPIRATION_EPOCH = "System-Expiration-Epoch"
SYSTEM_EXPIRATION_DURATION = "System-Expiration-Duration"
SYSTEM_EXPIRATION_TIMESTAMP = "System-Expiration-Timestamp"
SYSTEM_EXPIRATION_RFC3339 = "System-Expiration-RFC3339"


@pytest.mark.http_gate
@pytest.mark.http_put
class Test_http_system_header(ClusterTestBase):
    @pytest.fixture(scope="class")
    @allure.title("epoch_duration in seconds")
    def epoch_duration(self, default_wallet: WalletInfo) -> int:
        net_info = get_netmap_netinfo(
            wallet=default_wallet,
            endpoint=self.cluster.default_rpc_endpoint,
            shell=self.shell,
        )
        epoch_duration_in_blocks = net_info["epoch_duration"]
        time_per_block = net_info["time_per_block"]
        return int(epoch_duration_in_blocks * time_per_block)

    @allure.title("Return N-epoch count in minutes")
    def epoch_count_into_mins(self, epoch_duration: int, epoch: int) -> str:
        mins = epoch_duration * epoch / 60
        return f"{mins}m"

    @allure.title("Return future timestamp after N epochs are passed")
    def epoch_count_into_timestamp(self, epoch_duration: int, epoch: int, rfc3339: Optional[bool] = False) -> str:
        current_datetime = datetime.datetime.utcnow()
        epoch_count_in_seconds = epoch_duration * epoch
        future_datetime = current_datetime + datetime.timedelta(seconds=epoch_count_in_seconds)
        if rfc3339:
            return future_datetime.isoformat("T") + "Z"
        else:
            return str(calendar.timegm(future_datetime.timetuple()))

    @allure.title("Check if (header_output) Key=Value exists and equal in passed (header_to_find)")
    def check_key_value_presented_header(self, header_output: dict, header_to_find: dict) -> bool:
        header_att = header_output["header"]["attributes"]
        for key_to_check, val_to_check in header_to_find.items():
            if key_to_check not in header_att or val_to_check != header_att[key_to_check]:
                logger.info(f"Unable to find {key_to_check}: '{val_to_check}' in {header_att}")
                return False
        return True

    @allure.title(f"Validate that only {EXPIRATION_EPOCH_HEADER} exists in header and other headers are abesent")
    def validation_for_http_header_attr(self, head_info: dict, expected_epoch: int) -> None:
        # check that __SYSTEM__EXPIRATION_EPOCH attribute has corresponding epoch
        assert self.check_key_value_presented_header(
            head_info, {EXPIRATION_EPOCH_HEADER: str(expected_epoch)}
        ), f'Expected to find {EXPIRATION_EPOCH_HEADER}: {expected_epoch} in: {head_info["header"]["attributes"]}'
        # check that {EXPIRATION_EPOCH_HEADER} absents in header output
        assert not (
            self.check_key_value_presented_header(head_info, {EXPIRATION_DURATION_HEADER: ""})
        ), f"Only {EXPIRATION_EPOCH_HEADER} can be displayed in header attributes"
        # check that {EXPIRATION_TIMESTAMP_HEADER} absents in header output
        assert not (
            self.check_key_value_presented_header(head_info, {EXPIRATION_TIMESTAMP_HEADER: ""})
        ), f"Only {EXPIRATION_TIMESTAMP_HEADER} can be displayed in header attributes"
        # check that {EXPIRATION_EXPIRATION_RFC} absents in header output
        assert not (
            self.check_key_value_presented_header(head_info, {EXPIRATION_EXPIRATION_RFC: ""})
        ), f"Only {EXPIRATION_EXPIRATION_RFC} can be displayed in header attributes"

    @allure.title("Put / get / verify object and return head command result to invoker")
    def oid_header_info_for_object(self, default_wallet: WalletInfo, container: str, test_file: str, attributes: dict):
        oid = upload_via_http_gate_curl(
            cid=container,
            filepath=test_file,
            endpoint=self.cluster.default_http_gate_endpoint,
            headers=attr_into_str_header_curl(attributes),
        )
        verify_object_hash(
            oid=oid,
            file_name=test_file,
            wallet=default_wallet,
            cid=container,
            shell=self.shell,
            nodes=self.cluster.storage_nodes,
            request_node=self.cluster.cluster_nodes[0],
        )
        head = head_object(
            wallet=default_wallet,
            cid=container,
            oid=oid,
            shell=self.shell,
            endpoint=self.cluster.default_rpc_endpoint,
        )
        return oid, head

    @allure.title("[NEGATIVE] Put object with expired epoch")
    @requires_container(REP_2_1_2_PUBLIC)
    def test_unable_put_expired_epoch(self, container: str, simple_object_size: ObjectSize):
        headers = attr_into_str_header_curl({"System-Expiration-Epoch": str(get_epoch(self.shell, self.cluster) - 1)})
        file_path = generate_file(simple_object_size.value)
        with reporter.step("Put object using HTTP with attribute Expiration-Epoch where epoch is expired"):
            upload_via_http_gate_curl(
                cid=container,
                filepath=file_path,
                endpoint=self.cluster.default_http_gate_endpoint,
                headers=headers,
                error_pattern="must be greater than current epoch",
            )

    @allure.title("[NEGATIVE] Put object with negative System-Expiration-Duration")
    @requires_container(REP_2_1_2_PUBLIC)
    def test_unable_put_negative_duration(self, container: str, simple_object_size: ObjectSize):
        headers = attr_into_str_header_curl({"System-Expiration-Duration": "-1h"})
        file_path = generate_file(simple_object_size.value)
        with reporter.step("Put object using HTTP with attribute System-Expiration-Duration where duration is negative"):
            upload_via_http_gate_curl(
                cid=container,
                filepath=file_path,
                endpoint=self.cluster.default_http_gate_endpoint,
                headers=headers,
                error_pattern=f"{EXPIRATION_DURATION_HEADER} must be positive",
            )

    @allure.title("[NEGATIVE] Put object with System-Expiration-Timestamp value in the past")
    @requires_container(REP_2_1_2_PUBLIC)
    def test_unable_put_expired_timestamp(self, container: str, simple_object_size: ObjectSize):
        headers = attr_into_str_header_curl({"System-Expiration-Timestamp": "1635075727"})
        file_path = generate_file(simple_object_size.value)
        with reporter.step("Put object using HTTP with attribute System-Expiration-Timestamp where duration is in the past"):
            upload_via_http_gate_curl(
                cid=container,
                filepath=file_path,
                endpoint=self.cluster.default_http_gate_endpoint,
                headers=headers,
                error_pattern=f"{EXPIRATION_TIMESTAMP_HEADER} must be in the future",
            )

    @allure.title("[NEGATIVE] Put object using HTTP with attribute System-Expiration-RFC3339 where duration is in the past")
    @requires_container(REP_2_1_2_PUBLIC)
    def test_unable_put_expired_rfc(self, container: str, simple_object_size: ObjectSize):
        headers = attr_into_str_header_curl({"System-Expiration-RFC3339": "2021-11-22T09:55:49Z"})
        file_path = generate_file(simple_object_size.value)
        upload_via_http_gate_curl(
            cid=container,
            filepath=file_path,
            endpoint=self.cluster.default_http_gate_endpoint,
            headers=headers,
            error_pattern=f"{EXPIRATION_EXPIRATION_RFC} must be in the future",
        )

    @allure.title("Priority of attributes epoch>duration (obj_size={object_size})")
    @requires_container(REP_2_1_2_PUBLIC)
    def test_http_attr_priority_epoch_duration(
        self, default_wallet: WalletInfo, container: str, object_size: ObjectSize, epoch_duration: int
    ):
        self.tick_epoch()
        epoch_count = 1
        expected_epoch = get_epoch(self.shell, self.cluster) + epoch_count
        logger.info(
            f"epoch duration={epoch_duration}, current_epoch= {get_epoch(self.shell, self.cluster)} expected_epoch {expected_epoch}"
        )
        attributes = {SYSTEM_EXPIRATION_EPOCH: expected_epoch, SYSTEM_EXPIRATION_DURATION: "1m"}
        file_path = generate_file(object_size.value)
        with reporter.step(
            f"Put objects using HTTP with attributes and head command should display {EXPIRATION_EPOCH_HEADER}: {expected_epoch} attr"
        ):
            oid, head_info = self.oid_header_info_for_object(default_wallet, file_path, attributes, container)
            self.validation_for_http_header_attr(head_info=head_info, expected_epoch=expected_epoch)
        with reporter.step("Check that object becomes unavailable when epoch is expired"):
            for _ in range(0, epoch_count + 1):
                self.tick_epoch()
            assert (
                get_epoch(self.shell, self.cluster) == expected_epoch + 1
            ), f"Epochs should be equal: {get_epoch(self.shell, self.cluster)} != {expected_epoch + 1}"

            with reporter.step("Check object deleted because it expires-on epoch"):
                wait_for_epochs_align(self.shell, self.cluster)
                try_to_get_object_and_expect_error(
                    cid=container,
                    oid=oid,
                    node=self.cluster.cluster_nodes[0],
                    error_pattern="404 Not Found",
                )
                # check that object is not available via grpc
                with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                    get_object_from_random_node(default_wallet, container, oid, self.shell, self.cluster)

    @allure.title("Priority of attributes duration>timestamp (obj_size={object_size})")
    @requires_container(REP_2_1_2_PUBLIC)
    def test_http_attr_priority_dur_timestamp(
        self, default_wallet: WalletInfo, container: str, object_size: ObjectSize, epoch_duration: int
    ):
        self.tick_epoch()
        epoch_count = 2
        expected_epoch = get_epoch(self.shell, self.cluster) + epoch_count
        logger.info(
            f"epoch duration={epoch_duration}, current_epoch= {get_epoch(self.shell, self.cluster)} expected_epoch {expected_epoch}"
        )
        attributes = {
            SYSTEM_EXPIRATION_DURATION: self.epoch_count_into_mins(epoch_duration=epoch_duration, epoch=2),
            SYSTEM_EXPIRATION_TIMESTAMP: self.epoch_count_into_timestamp(epoch_duration=epoch_duration, epoch=1),
        }
        file_path = generate_file(object_size.value)
        with reporter.step(
            f"Put objects using HTTP with attributes and head command should display {EXPIRATION_EPOCH_HEADER}: {expected_epoch} attr"
        ):
            oid, head_info = self.oid_header_info_for_object(default_wallet, file_path, attributes, container)
            self.validation_for_http_header_attr(head_info=head_info, expected_epoch=expected_epoch)
        with reporter.step("Check that object becomes unavailable when epoch is expired"):
            for _ in range(0, epoch_count + 1):
                self.tick_epoch()
            assert (
                get_epoch(self.shell, self.cluster) == expected_epoch + 1
            ), f"Epochs should be equal: {get_epoch(self.shell, self.cluster)} != {expected_epoch + 1}"

            with reporter.step("Check object deleted because it expires-on epoch"):
                wait_for_epochs_align(self.shell, self.cluster)
                try_to_get_object_and_expect_error(
                    cid=container,
                    oid=oid,
                    node=self.cluster.cluster_nodes[0],
                    error_pattern="404 Not Found",
                )
                # check that object is not available via grpc
                with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                    get_object_from_random_node(default_wallet, container, oid, self.shell, self.cluster)

    @allure.title("Priority of attributes timestamp>Expiration-RFC (obj_size={object_size})")
    @requires_container(REP_2_1_2_PUBLIC)
    def test_http_attr_priority_timestamp_rfc(
        self, default_wallet: WalletInfo, container: str, object_size: ObjectSize, epoch_duration: int
    ):
        self.tick_epoch()
        epoch_count = 2
        expected_epoch = get_epoch(self.shell, self.cluster) + epoch_count
        logger.info(
            f"epoch duration={epoch_duration}, current_epoch= {get_epoch(self.shell, self.cluster)} expected_epoch {expected_epoch}"
        )
        attributes = {
            SYSTEM_EXPIRATION_TIMESTAMP: self.epoch_count_into_timestamp(epoch_duration=epoch_duration, epoch=2),
            SYSTEM_EXPIRATION_RFC3339: self.epoch_count_into_timestamp(epoch_duration=epoch_duration, epoch=1, rfc3339=True),
        }
        file_path = generate_file(object_size.value)
        with reporter.step(
            f"Put objects using HTTP with attributes and head command should display {EXPIRATION_EPOCH_HEADER}: {expected_epoch} attr"
        ):
            oid, head_info = self.oid_header_info_for_object(default_wallet, file_path, attributes, container)
            self.validation_for_http_header_attr(head_info=head_info, expected_epoch=expected_epoch)
        with reporter.step("Check that object becomes unavailable when epoch is expired"):
            for _ in range(0, epoch_count + 1):
                self.tick_epoch()
            assert (
                get_epoch(self.shell, self.cluster) == expected_epoch + 1
            ), f"Epochs should be equal: {get_epoch(self.shell, self.cluster)} != {expected_epoch + 1}"

            with reporter.step("Check object deleted because it expires-on epoch"):
                wait_for_epochs_align(self.shell, self.cluster)
                try_to_get_object_and_expect_error(
                    cid=container,
                    oid=oid,
                    node=self.cluster.cluster_nodes[0],
                    error_pattern="404 Not Found",
                )
                # check that object is not available via grpc
                with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                    get_object_from_random_node(default_wallet, container, oid, self.shell, self.cluster)

    @allure.title("Object should be deleted when expiration passed (obj_size={object_size})")
    @pytest.mark.parametrize(
        "object_size",
        # TODO: "complex" temporarly disabled for v0.37
        ["simple"],
        indirect=True,
    )
    @requires_container(REP_2_1_2_PUBLIC)
    def test_http_rfc_object_unavailable_after_expir(
        self, default_wallet: WalletInfo, container: str, object_size: ObjectSize, epoch_duration: int
    ):
        self.tick_epoch()
        epoch_count = 2
        expected_epoch = get_epoch(self.shell, self.cluster) + epoch_count
        logger.info(
            f"epoch duration={epoch_duration}, current_epoch= {get_epoch(self.shell, self.cluster)} expected_epoch {expected_epoch}"
        )
        attributes = {SYSTEM_EXPIRATION_RFC3339: self.epoch_count_into_timestamp(epoch_duration=epoch_duration, epoch=2, rfc3339=True)}
        file_path = generate_file(object_size.value)
        with reporter.step(
            f"Put objects using HTTP with attributes and head command should display {EXPIRATION_EPOCH_HEADER}: {expected_epoch} attr"
        ):
            oid, head_info = self.oid_header_info_for_object(default_wallet, file_path, attributes, container)
            self.validation_for_http_header_attr(head_info=head_info, expected_epoch=expected_epoch)
        with reporter.step("Check that object becomes unavailable when epoch is expired"):
            for _ in range(0, epoch_count + 1):
                self.tick_epoch()
            # check that {EXPIRATION_EXPIRATION_RFC} absents in header output
            assert (
                get_epoch(self.shell, self.cluster) == expected_epoch + 1
            ), f"Epochs should be equal: {get_epoch(self.shell, self.cluster)} != {expected_epoch + 1}"

            with reporter.step("Check object deleted because it expires-on epoch"):
                wait_for_epochs_align(self.shell, self.cluster)
                try_to_get_object_and_expect_error(
                    cid=container,
                    oid=oid,
                    node=self.cluster.cluster_nodes[0],
                    error_pattern="404 Not Found",
                )
                # check that object is not available via grpc
                with pytest.raises(Exception, match=OBJECT_NOT_FOUND):
                    get_object_from_random_node(default_wallet, container, oid, self.shell, self.cluster)
